#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
BackupMonitor.py

A script to notify the need for a backup of my personal data.
  
 Copyright (c) 2008-2009 Thibauld Nion
  
 Author: Thibauld Nion

  This program is free software; you can redistribute it and/or 
 modify it under the terms of the GNU General Public License as 
 published by the Free Software Foundation; either version 3 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA
"""

__author__="Thibauld Nion"
__license__="GPLv3"
__version__="1.6.2"


# These names refer to the rsync-incr-***.sh script where *** stands
# for one of the names below. They correspond to the 4 homonymous
# folders defined in the xdg-utils "standard" (so one should not worry
# about translating hem, xdg does that already).
DEFAULT_XDG_FOLDERS_TO_BACKUP=[
"Documents",
"Pictures",
"Music",
"Videos",
]


MAX_TIME = 2592000 # ie one month between backups
CHECK_TIMEOUT=  1000*60*60*24 # ie one check every day

# # Testing stuff
# CHECK_TIMEOUT=  1000*10 # ie one check every 10s
# MAX_TIME = 1

import os
import time
import subprocess as sp
import threading

import logging
import ConfigParser


import dbus


# ----------------------------------------------------------------------------
# Path configuration
# ----------------------------------------------------------------------------
#
# Define the folder where backups should be sotred (e.g. a symbolic
# link to a nice storage device)
BKP_FOLDER=os.path.expanduser("~/backup_destination")
BKP_LOG = os.path.expanduser("~/.backup_log")
# The prefix where the files have been put (in a system install that
# could be /usr/local for instance)
INSTALL_PREFIX=os.path.expanduser("~/.local")

# Place where default rsync scripts are stored
SCRIPTS_INSTALL_FOLDER = os.path.join(INSTALL_PREFIX,"share","backup-monitor",
                                      "scripts")
# Directory containing translations
LOCALISATION_LOCAL_PATH=os.path.join(INSTALL_PREFIX,"share","locale")
# Full path to the rsync script
BACKUP_MONITOR_RSYNC_SCRIPT = os.path.join(SCRIPTS_INSTALL_FOLDER,
                                           "rsync-incr-procedure.sh")
# Path to configuration file
USER_CONFIG_DIR = os.path.expanduser("~/.config/backup-monitor")
USER_SCRIPTS_FOLDER = os.path.join(USER_CONFIG_DIR,"user-scripts")
#
# ----------------------------------------------------------------------------





class BackupProcess(threading.Thread):
    """
    Grant access to the configuration and step by step action of the
    backup procedure. The whole backup can be run a thread.

    .. Warning:: Once created, a same instance won't take into account
                 any change in the configuration file.

    The most basic usage of this class would look like the
    following code (althouth you are encouraged to do
    something much more refined ;) )::

      bkp_process = BackupProcess(monitor)
      success = bkp_process.load_config()
      if success:
          bkp_process.start()

    It is also possible to call each backup step separately::
    
      bkp_process = BackupProcess(monitor)
      success = bkp_process.load_config()
      for itemName in bkp_process.get_backup_items():
          success = bkp_process.backup(itemName)
          if not success: break
          
    .. Warning:: When not using the start method, there is no guaranty
    that all backup items have actually been backed-up, this is left
    to the responsibility of the caller.
    """
    
    def __init__(self,monitor):
        threading.Thread.__init__(self)
        # keep link to the monitoring service that will coordinate
        # user interactions with this process
        self._monitor = monitor
        # default destination for backup
        self._bkp_dest_folder = BKP_FOLDER
        # default core script for backup
        self._bkp_core_script = BACKUP_MONITOR_RSYNC_SCRIPT
        # mapping backup 'items' vs corresponding method
        self._backup_items = {}
        # a hack that duplicates the list of keys of _backup_items,
        # but that will help control the order in which names are
        # accessed.
        self._backup_items_names = []
        # configuration placeholder (None until a valid configuration
        # has been loaded)
        self._config = None
        # indicate wether this thread should be paused
        self._pause_event = threading.Event()
        # indicate wether this thread should be killed
        self._kill_event = threading.Event()

    
    def get_pause_event(self):
        return self._pause_event

    def get_kill_event(self):
        return self._kill_event
        
        
    def load_config(self):
        """
        Load user/site configuration.

        Return true if no error was found in configuration.
        """
        config = ConfigParser.SafeConfigParser()
        # the config file will only exist if the user created one, let's
        # test it now.
        config_file_path=os.path.join(USER_CONFIG_DIR,"preferences.cfg")
        if os.access(config_file_path,os.R_OK):
            config.read(config_file_path)
        # check if user wants a specific destination folder
        if config.has_option("ChangeDefaults","DestinationFolder"):
            self._bkp_dest_folder=os.path.expanduser(config.get("ChangeDefaults",
                                                           "DestinationFolder"))
        # now register all items that should eventually be backed up
        itemsToBackup = {}
        itemsNamesList = []
        #---
        # Load config details about defaut folders and check if user
        # wants to disable backups for a given folder
        skip_list=[]
        if config.has_option("ChangeDefaults","SkipList"):
            skip_list=config.get("ChangeDefaults","SkipList").split(",")
        # only register non-skipped folders (obviously)
        for valuable_folder in DEFAULT_XDG_FOLDERS_TO_BACKUP:
            if valuable_folder in skip_list:
                logging.info("'%s' will be skipped per user request" % valuable_folder)
                continue
            else:
                itemsToBackup[valuable_folder] = self.backup_default_folder
                itemsNamesList.append(valuable_folder)
        #---
        # Load config details about the user-defined folders. To do
        # so: scan all config section, and look for those of type
        # "folder"
        count=0
        for sectionName in config.sections():
            if ( config.has_option(sectionName,"Type")
                 and config.get(sectionName,"Type")=="folder" ):
                if sectionName in itemsToBackup:
                    gobject.idle_add(self._monitor.show_error,_("Configuration error !"),
                                 _("""\
A user-defined folder is associated to the same name as a default folder:

%s""") % sectionName)
                    return False
                if config.has_option(sectionName,"FolderPath"):
                    # Preventive check (detect early that the indicated folder is wrong)
                    if not os.access(os.path.expanduser(config.get(sectionName,"FolderPath")),os.R_OK|os.X_OK):
                        gobject.idle_add(self._monitor.show_error,_("Configuration error !"),
                                         _("""\
The following user-defined folder cannot be accessed:

%s""") % sectionName)
                        return False
                    itemsToBackup[sectionName] = self.backup_user_folder
                    itemsNamesList.append(sectionName)
                    count+=1
                else:
                    gobject.idle_add(self._monitor.show_error,_("Configuration error !"),
                                 _("""\
Config section %s has 'folder' type but does not define a 'FolderPath'""") % sectionName)
                    return False
        if count==0:
            logging.info("No user-defined folders to backup")
        #---
        # Load config details about the user scripts. To do so: scan
        # all config section, and look for those of type "script"
        count=0
        for sectionName in config.sections():
            if ( config.has_option(sectionName,"Type")
                 and config.get(sectionName,"Type")=="script" ):
                if sectionName in itemsToBackup:
                    gobject.idle_add(self._monitor.show_error,_("Configuration error !"),
                                 _("""\
A user script is associated to the same name as a default folder or user-defined script:

%s""") % sectionName)
                    return False
                if config.has_option(sectionName,"ScriptName"):
                    itemsToBackup[sectionName] = self.launch_user_script
                    itemsNamesList.append(sectionName)
                    count+=1
                else:
                    gobject.idle_add(self._monitor.show_error,_("Configuration error !"),
                                 _("""\
Config section %s has 'script' type but does not define a 'ScriptName'""") % sectionName)
                    return False
        if count==0:
            logging.info("No user-defined scripts to launch")
        # up to this point we can remember about the config.
        self._config = config
        self._backup_items = itemsToBackup
        self._backup_items_names = itemsNamesList
        # so far, so good...
        return True



    
    def get_backup_items(self):
        """
        Get the list of names of all backup items.

        .. Hint:: call this after a successul call to load_config, or
                  you'll get an empty list.
        """
        # return self._backup_items.keys()
        # prefer returning the list where we control the order
        return self._backup_items_names
    
        
    def _is_backup_possible(self):
        """
        Check that comon required elements are available, actually
        checks:
        
        - destination folder
        - the core rsync script

        Other checks are left to each specific methods.

        .. Note:: it should be called at least at the begining of each backup method.
        """
        if not os.access(self._bkp_core_script,os.X_OK):
            gobject.idle_add(self._monitor.show_error,_("Cannot find the backup script"),
                                 _("""\
Please check that the following script exists and is executable:

%s""") % self._bkp_core_script)
            return False
        if not os.access(self._bkp_dest_folder,os.W_OK):
            gobject.idle_add(self._monitor.show_error,_("Impossible to write in the distant backup folder."),
                                       _("""\
Please check that you are able to write in the following directory:

%s""") % self._bkp_dest_folder)
            return False
        return True
        
    
    def _set_up_environment(self):
        """
        Set up the conventional set of environment variable that all
        scripts can rely on.
        """
        os.environ["BACKUP_MONITOR_RSYNC_SCRIPT"] = self._bkp_core_script
        os.environ["BACKUP_MONITOR_DESTINATION_FOLDER"] = self._bkp_dest_folder

    def _clean_up_environment(self):
        """
        Remove environment variables set up for the scripts.

        .. Warning:: assume that these variable have been set.

        .. See:: _set_up_environment
        """
        # cleanup environment
        del os.environ["BACKUP_MONITOR_RSYNC_SCRIPT"]
        del os.environ["BACKUP_MONITOR_DESTINATION_FOLDER"]
        

    def backup(self,name):
        """
        Backup an "item": this method will determine the "type" of the
        item (default folder, user-defined folder, or user-script) and
        launch the appropriate backup routine.
        
        .. Note:: This method is called when the thread "runs", and
        must communicate with gtk objects outside the thread via
        ``gobject.idle_add``
        """
        return self._backup_items[name](name)
        
        
    def backup_default_folder(self,name):
        """
        Perform the backup of one of the default directories.

        .. Note:: This method is called when the thread "runs", and
        must communicate with gtk objects outside the thread via
        ``gobject.idle_add``
        """
        # get information about the folder:
        if name not in self._backup_items:
            # that's more an internal error, let's just show info on
            # console and tell the caller that something's wrong.
            logging.info("Asking for the backup of a forbidden folder (according to config): '%s'" % name)
            return False
        # sanity check: is everything in place ?
        if not self._is_backup_possible():
            return False
        # also check presence of the "xdg" specific script
        bkp_script_for_xdg = os.path.join(SCRIPTS_INSTALL_FOLDER,
                                          "rsync-incr-xdg-dir.sh")
        if not os.access(bkp_script_for_xdg,os.X_OK):
            gobject.idle_add(self._monitor.show_error,
                             _("Cannot find the default backup script"),
                             _("""\
Please check that the following script exists and is executable:

%s""") % bkp_script_for_xdg)
            return False
        # everything's ok ? well then... launch the actual backup
        self._set_up_environment()
        try:
            # create the command needed to launch our core rsync script
            bkp_cmd = "%s %s" % (bkp_script_for_xdg,name.upper())
            logging.info("Launching %s" % bkp_cmd)
            # generate stdout and stderr redirection option
            redirection = " >> %s 2>&1" % BKP_LOG
            estatus = sp.call( bkp_cmd + redirection, shell=True)
            logging.info("Return status: %s" % estatus)
            if estatus > 0:
                gobject.idle_add(self._monitor.show_error_and_log,
                                 _("Error during backup of %s") % name,
                                 _("Close this dialog to see the error log."))
                return False
            else:
                return True
        finally:
            self._clean_up_environment()
        

    def backup_user_folder(self,name):
        """
        Perform the backup of a user defined directory.

        .. Note:: This method is called when the thread "runs", and
        must communicate with gtk objects outside the thread via
        ``gobject.idle_add``
        """
        # get information about the folder:
        if name not in self._backup_items:
            # that's more an internal error, let's just show info on
            # console and tell the caller that something's wrong.
            logging.info("Asking for the backup that was never requested: '%s'" % name)
            return False
        if self._config.has_option(name,"FolderPath"):
            folderFullPath = os.path.expanduser(self._config.get(name,"FolderPath"))            
        else:
            # that's also more an internal error, let's just show info on
            # console and tell the caller that something's wrong.
            logging.info("backup_user_folder has been wrongly called for a non-user-folder backup")
            return False
        if not os.access(folderFullPath,os.R_OK):
            gobject.idle_add(self._monitor.show_error,
                             _("Cannot access the directory to backup"),
                             _("Please check your configuration for '%s'")%name)
            return False
        # sanity check: is everything in place ?
        if not self._is_backup_possible():
            return False
        # everything's ok ? well then... launch the actual backup
        self._set_up_environment()
        try: 
            bkp_cmd = "%s %s %s" % (self._bkp_core_script,folderFullPath,self._bkp_dest_folder)
            logging.info("Launching %s" % bkp_cmd)
            # generate stdout and stderr redirection option
            redirection = " >> %s 2>&1" % BKP_LOG
            estatus = sp.call( bkp_cmd + redirection, shell=True)
            logging.info("Return status: %s" % estatus)
            if estatus > 0:
                gobject.idle_add(self._monitor.show_error_and_log,
                                 _("Error during backup of %s") % name,
                                 _("Close this dialog to see the error log."))
                return False
            else:
                return True
        finally:
            self._clean_up_environment()


    def launch_user_script(self,name):
        """
        Launch a script indicated in the config file.

        .. Note:: This method is called when the thread "runs", and
        must communicate with gtk objects outside the thread via
        ``gobject.idle_add``
        """
        # get information about the folder:
        if name not in self._backup_items:
            # that's more an internal error, let's just show info on
            # console and tell the caller that something's wrong.
            logging.info("Asking for the backup that was never requested: '%s'" % name)
            return False
        if self._config.has_option(name,"ScriptName"):
            scriptFullPath = os.path.join(USER_SCRIPTS_FOLDER,
                                          self._config.get(name,"ScriptName"))
        else:
            # that's also more an internal error, let's just show info on
            # console and tell the caller that something's wrong.
            logging.info("launch_user_script has been wrongly called for something that is not a user-defined script")
            return False
        logging.info("Checking executability of script: %s" % scriptFullPath)
        if not os.access(scriptFullPath,os.X_OK):
            gobject.idle_add(self._monitor.show_error,
                             _("Cannot execute user-defined script"),
                             _("Please check your configuration for '%s'")%name)
            return False
        # sanity check: is everything in place ?
        if not self._is_backup_possible():
            return False
        # everything's ok ? well then... launch the actual backup
        self._set_up_environment()
        try: 
            logging.info("Launching %s" % scriptFullPath)
            # generate stdout and stderr redirection option
            redirection = " >> %s 2>&1" % BKP_LOG
            estatus = sp.call( scriptFullPath + redirection, shell=True)
            logging.info("Return status: %s" % estatus)
            if estatus > 0:
                gobject.idle_add(self._monitor.show_error_and_log,
                                 _("Error during backup of %s") % name,
                                 _("Close this dialog to see the error log."))
                pid = sp.Popen(["xdg-open", BKP_LOG]).pid
                return False
            else:
                return True
        finally:
            self._clean_up_environment()

    def run(self):
        """
        To be run as a thread.

        .. Warning:: each backup operation is "atomic", and this
        thread will only pause or stop in between those operations.
        
        .. Note:: All method called from here (ie when the thread
        "runs") must communicate with gtk objects outside the thread
        via ``gobject.idle_add``
        """
        # check if config has already been loaded
        if self._config is None:
            success = self.load_config()
            if not success:
                return gobject.idle_add(self._monitor.backup_finished,False)
        for idx,itemName in enumerate(self.get_backup_items()):
            while self._pause_event.isSet():
                pass
            if self._kill_event.isSet():
                # no signal here, kill is an external event, the
                # caller should be aware of it.
                return
	    gobject.idle_add(self._monitor.set_item_as_wip,idx)
	    success = self.backup(itemName)
	    if success:
		gobject.idle_add(self._monitor.set_item_as_done,idx)
		continue
	    else:
		break
        gobject.idle_add(self._monitor.backup_finished,success)


import gobject
import gtk
import pango

class GtkBackupItemList:
    """
    Display list of items with some status icons.

    .. Warning:: this class is in coupling with the GtkMonitor, it's
    not really goog for design, but does the job here...
    """
    status_icons = {
        "todo": gtk.STOCK_JUMP_TO,
        "done": gtk.STOCK_APPLY,
        "wip" : gtk.STOCK_EXECUTE,
        }
    
    
    def __init__(self,monitor,itemList,keepBelow=False):
        # Following instruction from
        # http://faq.pygtk.org/index.py?req=show&file=faq20.006.htp
        #
        # This is to allow using threads to update the GUI
        gobject.threads_init()
        
	# create a new window
	self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
	self.window.set_title(_("BackupMonitor"))
	# the monitor UI
	self.monitor = monitor
	# Create a list store model
	self._init_store(itemList)
	# Create the view
	self._init_tree_view()
	# Put it in a scroll bar
	self.scrolledwindow = gtk.ScrolledWindow()
	self.scrolledwindow.set_policy(gtk.POLICY_AUTOMATIC,
				       gtk.POLICY_AUTOMATIC)
	self.scrolledwindow.add(self.treeview)
	
	# Create upper text area
	self._init_txt_area()
	# Create buttons
	self._init_bottom_buttons()
	
	# This packs the buttons and the treeview.
	self.mainBox = gtk.VBox()
	self.mainBox.pack_start(self.topBox,expand=False)
	self.mainBox.pack_start(self.scrolledwindow,expand=True)
	self.mainBox.pack_start(self.downBox,expand=False)
	self.window.add(self.mainBox)

        gtk.window_set_default_icon_name("backup-monitor")
	# When the window is given the "delete_event" signal (close)
        self.window.connect("delete_event", self.cancel)
        
	# Ask for a good initial size
	self._set_good_window_size(len(itemList))
        # A little flag/hack to detect when the window is de-iconified
        # for the first time after the launch
        self._iconified=False
	# The final step is to display this newly created window.
	self.window.show_all()
        if keepBelow:
            # But signals itself as important anyway
            self.window.set_urgency_hint(True)
            # Make sure it does not stell the focus
            self.window.iconify()
            # should be called at the end because it gets some focus
            # during the "show" apparently
            self.window.connect("window-state-event", self.on_focus)


    def on_focus(self, widget, event):
        """
        When the window gets in focus, detect if it's the first time
        that it is de-iconified.
        """
        if (event.changed_mask & gtk.gdk.WINDOW_STATE_ICONIFIED):
            if not self._iconified:
                self._iconified = True
            else:
                self.window.set_urgency_hint(False)

    
    def go_on_top(self):
        """
        Show window on top of others.
        """
        self.window.present()

    def destroy(self):
        """
        Get rid of all the GUI elements.
        """
        self.window.destroy()
    
    def _init_store(self,itemList):
	"""
	Create the liststore model.
	"""
	# In the model we have:
	# [0]: status (done,wip,todo)
	# [1]: funny icon (linked to status)
	# [2]: text description
	self.liststore = gtk.ListStore(str, str, str)
	# Fill in the list store
	for itemName in itemList:
	    self.liststore.append(["todo", 
				   self.status_icons["todo"],
				   itemName])


    def set_item_as_done(self,itemIdx):
        """
        Change an item's status to done
        """
        self.liststore[itemIdx][0] = "done"
        self.liststore[itemIdx][1] = self.status_icons["done"]
        self.window.queue_draw()
        
    def set_item_as_wip(self,itemIdx):
        """
        Change an item status to wip (work in progress)
        """
        self.liststore[itemIdx][0] = "wip"
        self.liststore[itemIdx][1] = self.status_icons["wip"]
        self.window.queue_draw()

    def set_backup_has_started(self):
        """
        Indicate on the GUI that the backup has started.
        """
        self.btnApply.set_label("Running...")
        self.btnCancel.set_use_stock(True)
        self.btnCancel.set_label("gtk-stop")
        
        
        
    def set_backup_all_done(self):
        """
        Indicate on the GUI that backup is finished.
        """
        self.btnApply.set_label("All done.")
        self.downBox.set_sensitive(False)
    
    def _set_good_window_size(self,nb_items):
        """
	Try to set a good initial size for the window
	"""
	TXT_WRAP_LENGTH = 40
	pctx = self.treeview.get_pango_context()
	l = pango.Layout(pctx)
	l.set_text("".join(["a" for i in range(TXT_WRAP_LENGTH)]))
	txt_width, txt_height = l.get_pixel_size()
	# get a little more space for the check box and icon
	width = min(int(txt_width),
		    int(gtk.gdk.screen_width()*2./3.) )
	# approximate the vertical height by considering that some
	# tasks have to be put on 2 or more lines
	height = min( int(2* nb_items * txt_height),
		      int(gtk.gdk.screen_height()*1./2.) )
	self.treeview.set_size_request(width,height)
	

    def _init_txt_area(self):
	"""
	Description text
	"""
	message = gtk.Label()
	message.set_markup("\n<b>" + _("Your data must be backed up !") + "</b>"
                           + "\n\n" + _("""\
No backup has been made recently. 
Please click on the 'Apply' button to launch the backup procedure on the following items.""") + "\n\n" )
        message.set_line_wrap(True)
        self.topBox = gtk.HBox()
        self.topBox.pack_start(message,expand=True,padding=10)
        
	     
    def _init_bottom_buttons(self):
        """
        Layout the bottom buttons
        """
        # Creates validation/cancellation buttons.
        self.btnCancel = gtk.Button(stock=gtk.STOCK_CANCEL)
        self.btnApply = gtk.Button(stock=gtk.STOCK_APPLY)
        # connect events
        self.btnCancel.connect("clicked", self.cancel, None)
        self.btnApply.connect("clicked", self.apply, None)
        # Layout the lower buttons
        self.downBox = gtk.HBox()
        self.downBox.pack_start(self.btnCancel)
        self.downBox.pack_end(self.btnApply)
        

    def _init_tree_view(self):
        """
        Create the tree view that will display the liststore.
        """		
        self.treeview = gtk.TreeView(self.liststore)
        # Set some properties
        self.treeview.set_enable_search(True)
        self.treeview.set_search_column(2)
        
        # create a CellRenderers to render the data
        cellIcon = gtk.CellRendererPixbuf()
        cellName = gtk.CellRendererText()
	 
        # Create the TreeViewColumns to display the data
        tvcolumnStatus = gtk.TreeViewColumn(_('Status'),cellIcon)
        tvcolumnName = gtk.TreeViewColumn(_('Backup Item'),cellName)
	 
        # Add columns to treeview
        self.treeview.append_column(tvcolumnStatus)
        self.treeview.append_column(tvcolumnName)
        # Link the content of the columns with the content of the
        # store
        tvcolumnStatus.set_attributes(cellIcon, stock_id=1)
        tvcolumnName.set_attributes(cellName, markup=2)


    def set_sensitive(self,sense):
        """
        Make the displayed window sensitive or not.

        Also change the cursor to a watch.
        """
        self.window.set_sensitive(sense)
        watch = gtk.gdk.Cursor(gtk.gdk.WATCH)
        self.window.window.set_cursor(watch)
        gtk.gdk.flush()
        
        
    def apply(self,widget,data=None):
        """
        Apply and save the modifications of the tasks
        """
        self.monitor.start_backup()
        

    def cancel(self,widget,data=None):
        """
        Do not apply the done/not done settings
        """
        self.monitor.abort_request()

    
        

        
class GtkMonitor:
    """
    A simplistic class that handles the display of the start dialog.
    """
    
    def __init__(self,manager):
        """
        Init with predefind text
        """
        # the message dialog
        self.d = None
        self._generate_start_dialog()
        self.waiting = False
        self.manager = manager
        self.backup_process = None
        self.backup_process_pause = None
        self.backup_process_kill = None
        self.status_icon = None
        self.backup_window = None        
        
    def show(self):
        """
        Display youself to the user..
        """
        # create a status icon
        # self.status_icon = gtk.status_icon_new_from_stock(gtk.STOCK_SAVE)
        self.status_icon = gtk.status_icon_new_from_icon_name("backup-monitor")
        self.status_icon.set_tooltip(_("Click here to backup your data"))
        # allow for the user to launch the procedure
        self.status_icon.connect("activate", self.show_backup_window)
        self.show_backup_window(None)
        
    def _generate_start_dialog(self):
        """
        Generate the message dialog
        """
        self.waiting = False
        self.d = gtk.MessageDialog(type=gtk.MESSAGE_WARNING,
                                   flags=gtk.DIALOG_MODAL,
                                   buttons=gtk.BUTTONS_OK_CANCEL, 
                                   message_format=_("Launch the backup procedure ?"))
        self.d.format_secondary_text(_("Make sure the storage device for your backup is connected."))

    def _generate_waiting_dialog(self):
        """
        Generate the message dialog
        """
        self.waiting = True
        self.d = gtk.MessageDialog(type=gtk.MESSAGE_INFO,
                                   flags=gtk.DIALOG_MODAL,
                                   buttons=gtk.BUTTONS_OK, 
                                   message_format=_("The backup procedure is running."))
        self.d.format_secondary_text(_("Please concentrate on interesting things, this procedure is automatic !."))    
    
    def show_backup_window(self,status_icon):
        """
        Show a window explaining the need for a backup and allowing the
        user to launch it.

        If status_icon is None, then assume that it is not launched
        from status_icon, and ask the window to stay below existing
        apps' windows.
        """
        # self.d.connect("response",self.obey_user)
        # if not self.waiting:
        #     return self.d.show()
        # else:
        #     return
        
        # decide wether the window should be shown above others or
        # below
        if status_icon is None:
            keepBelow=True
        else:
            # event coming from a status icon
            if self.backup_window is None:
                # let's keep on an create one
                keepBelow=False
            else:
                # hey we already have a window, just display it on top
                return self.backup_window.go_on_top()
        # create the process now to get the list of items
        bkp_process = BackupProcess(self)
        success = bkp_process.load_config()
        if success:
            self.backup_process = bkp_process
            self.backup_process_pause = bkp_process.get_pause_event()
            self.backup_process_kill  = bkp_process.get_kill_event()
            # we have enough info to show the window
            self.backup_window = GtkBackupItemList(self,
                                                   self.backup_process.get_backup_items(),
                                                   keepBelow)
        else:
            # wait for the user to fix his/her config and get back to
            # us again (through the status icon for instance)
            return
        
    def _del_backup_window(self, widget=None, data=None):
        """
        Destroy backup window and remove any reference.

        Accepts args so that it can be connected to an event.
        """
        if self.backup_window is not None:
            self.backup_window.destroy()
            self.backup_window = None
            
            
    def start_backup(self):
        """
        Show preliminary (and important) dialog asking to check for
        connection to the backup device.
        """
        self.d.connect("response",self.obey_user)
        return self.d.show()
    
    def obey_user(self,d,response_id):
        """
        react to the user response to the dialog
        """
        d.destroy()
        if self.waiting:
            # regenerate a new dialog
            self._generate_waiting_dialog()
            return
        resp = response_id
        if resp==gtk.RESPONSE_OK:
            # regenerate a new dialog
            self._generate_waiting_dialog()
            self.status_icon.set_tooltip(_("Data backup running..."))
            # launch the backup procedure (that must handle its own
            # messages and dialogs)
            self.backup_process.start()
            self.backup_window.set_backup_has_started()
        if resp==gtk.RESPONSE_CANCEL:
            # regenerate a new dialog
            return self._generate_start_dialog()
                

    def abort_request(self):
        """
        Request for the whole backup to be aborted
        """
        if self.backup_process is None:
            # the backup was not launched, no need for an additional
            # dialog
            return self.abort_request_answer(None,gtk.RESPONSE_YES)
        if self.backup_process.isAlive():
            self.backup_process_pause.set()
        else:
            # the backup was not launched, no need for an additional
            # dialog
            return self.abort_request_answer(None,gtk.RESPONSE_YES)
        abortDialog = gtk.MessageDialog(type=gtk.MESSAGE_QUESTION,
                                        flags=gtk.DIALOG_MODAL,
                                        buttons=gtk.BUTTONS_YES_NO,
                                        message_format=_("Abort the backup ?"))
        abortDialog.format_secondary_text(
            _("""\
Click No to continue with the backup.

Click Yes and the backup will stop as soon as the currently running task is finished.
"""))
        
        abortDialog.connect("response",self.abort_request_answer)
        abortDialog.run()

    def abort_request_answer(self,d,response_id):
        """
        When the user clicked on one of the button of the abort dialog.

        .. Note:: we are accepting d to be None.
        """
        if d is not None:
            d.destroy()
        if response_id == gtk.RESPONSE_YES:
            if self.backup_process is not None:
                # if a backup process has been create ask it to stop
                self.backup_process_kill.set()
                self.backup_process_pause.clear()
                self.backup_window.set_sensitive(False)
                while self.backup_process.isAlive():
                    # wait for the thread to actually stop
                    self.backup_process.join()
            # another message to tell the user about the status icon
            seeYouLaterDialog = gtk.MessageDialog(type=gtk.MESSAGE_INFO,
                                                  flags=gtk.DIALOG_MODAL,
                                                  buttons=gtk.BUTTONS_OK, 
                                                  message_format=_("See you later !"))
            seeYouLaterDialog.format_secondary_text(
                _("""\
BackupMonitor will sleep now.

To start again, click on BackupMonitor's status icon."""))
            seeYouLaterDialog.run()
            seeYouLaterDialog.destroy()          
            self._cleanup_before_sleep()
        else:
            # by default resume the backup
            if self.backup_process is not None:
                self.backup_process_pause.clear()


    def set_item_as_done(self,itemIdx):
        """
        Indicate that a backup item is done.

        Item is defined by it's index in the list returned by
        BackupProcess.get_backup_items
        """
        self.backup_window.set_item_as_done(itemIdx)
        
    def set_item_as_wip(self,itemIdx):
        """
        Indicate that an item is being backed up.
        
        Item is defined by it's index in the list returned by
        BackupProcess.get_backup_items
        """
        self.backup_window.set_item_as_wip(itemIdx)
    
    def backup_finished(self,success):
        """
        To be called when backup process is finished with a status
        information: success should be True if everythng is fine, and
        false else.
        """
        if success:
            self.backup_window.set_backup_all_done()
            successDialog = gtk.MessageDialog(type=gtk.MESSAGE_INFO,
                                              flags=gtk.DIALOG_MODAL,
                                              buttons=gtk.BUTTONS_OK, 
                                              message_format=_("Backup successful !"))
            successDialog.format_secondary_text(
                _("Click OK to let BackupMonitor sleep until next time."))
            successDialog.run()
            successDialog.destroy()
            self._cleanup_before_sleep()
            self._get_back_under_manager_command()
        else:
            seeYouLaterDialog = gtk.MessageDialog(type=gtk.MESSAGE_INFO,
                                                  flags=gtk.DIALOG_MODAL,
                                                  buttons=gtk.BUTTONS_OK, 
                                                  message_format=_("See you later !"))
            seeYouLaterDialog.format_secondary_text(
                _("""\
An error happened but BackupMonitor remains at your service.

When the error is fixed, click on BackupMonitor's status icon to try again."""))
            seeYouLaterDialog.run()
            seeYouLaterDialog.destroy()
            self._cleanup_before_sleep()
            self.status_icon.set_from_icon_name("backup-monitor")
            self.status_icon.set_tooltip(_("Click here to backup your data"))


    def _cleanup_before_sleep(self):
        """
        Cleanup following items:
        - backup_process
        - backup_window

        .. Note:: Don't hide the statuc icon.
        """
        self._del_backup_window()
        self.backup_process = None
        self.backup_process_pause = None
        self.backup_process_kill = None
        return self._generate_start_dialog()
    
    def _get_back_under_manager_command(self):
        """
        Stop acting by its own, and let the upper-level manager
        request any new action.
        """
        self.status_icon.set_visible(False)
        if self.manager.is_set_for_one_shot_only():
            return self.manager.quit()
        else:
            # run the backup monitoring function every day
            gobject.timeout_add(CHECK_TIMEOUT, self.manager.check_backup_need) 
            # save current date
            return self.manager.stamp_backup_date()
        
    
    def show_error(self,summary,additional_text=None):
        """
        Display an error of the backup procedure
        """
        d = gtk.MessageDialog(type=gtk.MESSAGE_ERROR,
                              buttons=gtk.BUTTONS_OK, 
                              message_format=summary)
        if additional_text is not None:
            d.format_secondary_text(additional_text)
        d.run()
        d.destroy()


    def show_error_and_log(self,summary,additional_text=None):
        """
        Display an error (like show_error), but also display the full log.
        """
        self.show_error(summary,additional_text)
        # use the system's prefered text tool to display the log.
        pid = sp.Popen(["xdg-open", BKP_LOG]).pid
        
        
    def show_info(self,summary,additional_text=None):
        """
        Display some info about the backup procedure
        """
        # if additional_text is None:
        #     additional_text=""
        # self.manager.notify(summary,additional_text)
        logging.info("%s:\n\t%s" % (summary, additional_text))


        
    def notify(self,summary,additional_text="",timeout=-1):
        """
        Create a notification.
        """
        # use notification via dbus
        bus = dbus.SessionBus()
        notif = bus.get_object("org.freedesktop.Notifications",
                               "/org/freedesktop/Notifications")
        notif_interface = dbus.Interface(notif,
                                         "org.freedesktop.Notifications")
        notif_interface.Notify("Backup Notification",0,
                               "backup-monitor", summary,
                               additional_text,[],
                               {},timeout)

class BackupManager:
    """
    Class gathering the most top level function and which decide
    wether launching or not the backup procedure.
    """
    
    STAMP_FILEPATH = os.path.expanduser("~/.backup_stamp")

    def __init__(self):
        """
        Init some variables changing the overall behaviour.
        """
        # flag to know that no monitoring should be made
        self.oneShot = False
        # flag to know that the script is run only for test.
        self.testRun = False
        # flag to indicate if we should ignore the stamp and run a backup anyway
        self.forcedBackup = False
        # UI that will be used
        self.monitorUI = GtkMonitor(self) 
        

    def force_backup(self):
        """
        Force backup (not stamp)
        """
        self.forcedBackup = True

    def is_backup_forced(self):
        """
        Force backup (not stamp)
        """
        return self.forcedBackup

    def set_for_one_shot_only(self):
        """
        One shot (not stamp, not recurrence)
        """
        self.oneShot = True

    def is_set_for_one_shot_only(self):
        """
        One shot (not stamp, not recurrence)
        """
        return self.oneShot

    def set_for_a_test_run(self):
        """
        Test run: only perform a dry run.
        """
        self.testRun = True
        global BACKUP_MONITOR_RSYNC_SCRIPT
        # Full path to the rsync script
        BACKUP_MONITOR_RSYNC_SCRIPT = os.path.join(SCRIPTS_INSTALL_FOLDER,
                                                   "rsync-incr-procedure-dry.sh")
        
                
    def is_set_for_a_test_run(self):
        """
        Test run
        """
        return self.testRun
        
    def quit(self):
        """
        Quit the application
        """
        gtk.main_quit()

    def check_backup_need(self):
        """
        Check the need for backup.

        Return True by default and if no error is encoutered. If for
        some reason the function shuld not be called again (user
        configuration or error cases), then it will return False. This
        is so that it can be included in timeout function of simple
        loops
        """
        
        if not os.access(self.STAMP_FILEPATH,os.W_OK):
            try:
                f = open(self.STAMP_FILEPATH,"w")
                f.write("0")
                f.close()
            except:
                self.monitorUI.show_error(_("ERROR with the date stamp !"),
                                          _("""\
Impossible to create the following file needed to keep track of when the latest backup was done:

%s

Please check that this program is running with the proper credentials.""") % self.STAMP_FILEPATH)
                return False
        f = open(self.STAMP_FILEPATH,"r")
        r = f.read()
        
        logging.info( "Content of the date stamp: %s" % r )
        latest_epoch_time = int(r)
        f.close()
        current_epoch_time = int(time.time()) 
        elapsed_time = current_epoch_time-latest_epoch_time
        duration_as_months = elapsed_time/MAX_TIME
        if duration_as_months>=1 or self.forcedBackup:            
            self.ask_for_backup()
            return False
        else:
            logging.info("Latest backup recent enough: nothing to do.")
            if self.oneShot:
                logging.info("One shot process, stop here.")
                return False
            else:
                # keep on being called by the timeout function
                return True
    
    def stamp_backup_date(self):
        """
        Write the stamp file to remember the date of the latest backup.
        """
        if not self.oneShot:
            current_epoch_time = int(time.time()) 
            f = open(self.STAMP_FILEPATH,"w")
            f.write("%d"%current_epoch_time)
            f.close()
        else:
            pass
        
    
    def ask_for_backup(self):
        """
        Monitor the needs for backup (function whose execution is to
        be programmed via a timeout)
        """
        self.monitorUI.show()
        return False

        

if __name__=="__main__":

    from optparse import OptionParser
    
    sleepDuration = 2*60
    onlyOnce=False
    testRun=False

    parser = OptionParser("usage %prog [options]",
                          version="%prog " +  str(__version__),
                          description="Starts monitoring the need for backups. This program is best called at the begining of a user's desktop session. It is possible to change the default behaviour using the options below, but they are mainly intended for use by developers.")
    parser.add_option("-n","--now",
                      action="store_true",dest="now",default=False,
                      help="dont wait before actually launching the program.")
    parser.add_option("-v","--verbose",
                      action="store_true",dest="verbose",default=False,
                      help="talk to me pleaaaase.")
    parser.add_option("-o","--once",
                      action="store_true",dest="once",default=False,
                      help="launch the backup procedure only once (do not stay in background).")
    parser.add_option("-f","--force",
                      action="store_true",dest="force",default=False,
                      help="force a backup to happen (ignoring the date of latest run)")
    parser.add_option("-t","--test",
                      action="store_true",dest="test",default=False,
                      help="only print what should be done and perform an rsync dry run.")

    (options,args) = parser.parse_args()

    if len(args)>0:
        parser.error("not handling additional parameters like: '%s'" % " ".join(args))
    # for specific uses
    if options.now:
        sleepDuration = 0

    # set up the logging to the "global" log file
    logging.basicConfig(filename=os.path.expanduser(BKP_LOG),filemode='w',level=logging.INFO)
    if options.verbose:
        # define a Handler which writes INFO messages or higher to the sys.stderr
        #
        # (copied from http://docs.python.org/library/logging.html)
        console = logging.StreamHandler()
        console.setLevel(logging.INFO)
        # set a format which is simpler for console use
        formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')
        # tell the handler to use this format
        console.setFormatter(formatter)
        # add the handler to the root logger
        logging.getLogger('').addHandler(console)
        
    # take into account cli options.
    logging.info("Time before activation: %s" % sleepDuration)
    time.sleep(sleepDuration)
    
    # ----------------------------------------------------------------------------
    # Localisation stuff
    # ----------------------------------------------------------------------------
    # 
    import gettext
    # get the current locale
    import locale
    if locale.getdefaultlocale()[0] is None:    
        logging.info("No default locale found, assuming that messages don't have to be translated.")
        # if the local is not suppoted define the _() function anyway
        _ = lambda text:text
    else:
        try:
            LOCALISATION_LANG = locale.getdefaultlocale()[0][:2]
            LOCALISATION_CUR_LANG = gettext.translation('backup-monitor',
                                                        localedir=LOCALISATION_LOCAL_PATH, 
                                                        languages=[LOCALISATION_LANG])
            LOCALISATION_CUR_LANG.install()
        except IOError,e:
            logging.info("Couldn't load the translation corresponding to user's locale: %s" % LOCALISATION_CUR_LANG)
            # if the local is not suppoted define the _() function anyway
            _ = lambda text:text
     
    #
    # ----------------------------------------------------------------------------

    
    manager =  BackupManager()
    if options.once:
        manager.set_for_one_shot_only()
    if options.test:
        manager.set_for_a_test_run()
    if options.force:
        manager.force_backup()
    # anyway check backup needs right away
    manager.check_backup_need()
    
    # All PyGTK applications must have a gtk.main(). Control ends here
    # and waits for an event to occur (like a key press or mouse event).
    gtk.main()
