#!/bin/bash

# rsync-based script for incremental backup
#
# Author: Thibauld Nion
# License: Public Domain
#
#
# Perform an incremental backup so that the 'current' folder will
# mimic the content of the actual folder we are backuping.  Stuff not
# existing anymore on the local directory wil be copied to a BACKUPDIR
# folder (that is named after the date of the backup ie 2008-03-22 if
# this is the date when the script is run).
#
# The variable are supposed to be set in specific files that will
# call the present script aftewards.
#
# Original idea taken from:
# http://samba.anu.edu.au/rsync/examples.html
#
########################################################################


# The path to the directory to be backed up
BDIR_FULL_PATH=$1
# The destination of the backup
BDEST_BASE=$2
# The path to a file listing the files to be excluded from the backup
# (with wildcards)
EXCLUDES=$3

echo $BDEST_BASE
if [ ! -d $BDIR_FULL_PATH ]; then
    echo "ERROR ! Cannot find directory to back up: " $BDIR_FULL_PATH >&2
    exit 1
fi;

if [ "$BDEST_BASE" = "" ]; then
    echo "ERROR: no backup destination given." >&2
    exit 1
fi;

BDIR_MAIN_PREFIX=`dirname $BDIR_FULL_PATH`
BDIR_MAIN_NAME=`basename $BDIR_FULL_PATH`
# echo $BDIR_MAIN_PREFIX
# echo $BDIR_MAIN_NAME

echo "Starting backup procedure for:"
echo $BDIR_FULL_PATH 
echo "(backed up to " $BDEST_BASE ")"

if [ ! -d $BDEST_BASE ]; then
    echo "ERROR ! Destination of the backup is not an existing directory." >&2
    exit 1
fi;

# the destination of the backup
BDEST=$BDEST_BASE/$BDIR_MAIN_NAME
install -d $BDEST

BACKUPDIR=`date +%F`
OPTS="-a -k --no-g --no-p --no-o --protect-args --safe-links --force --delete --backup --backup-dir=$BDEST/$BACKUPDIR"
if [ "$EXCLUDES" != "" ]; then 
    OPTS="$OPTS	 --exclude-from=$EXCLUDES"
fi;
export PATH=$PATH:/bin:/usr/bin:/usr/local/bin

# only keep the two latest diffs
echo "Cleaning directories that are too old"
count=0
for f in `ls $BDEST`; do
    # in case the script is run twice the same date don't take
    # into account the newly created directory.
    if [ $BACKUPDIR != $f ]
    then
	let "count += 1" 
    fi
done;
cleanedcount=0
for f in `ls $BDEST`; do
    # suppress directories by alphabetical order (oldest
    # first), until we arrive at count<2
    if [ "2" -lt $count ]
    then
	let "count -= 1"
	echo "Removing deprecated directory: " $f
	rm -rf $BDEST/$f
	let "cleanedcount += 1"
    fi
done;
echo "	" $cleanedcount "directories cleaned"

# perform the backup
echo "Creating the backup copy"
pushd $BDIR_FULL_PATH  
rsync $OPTS ./ $BDEST/current
if [ $? -ne 0 ]; then
    popd
    echo "ERROR ! rsync command sent an error" >&2
    exit 1
fi
popd
echo "... done."

